/**
 * rclockd.cpp
 *
 * rclockd main program file.
 *
 * (c) 2008 John Sennesael
 * See the COPYING file for license information.
 *
 */

// global includes ============================================================

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <termios.h>
#include <unistd.h>
#include <stdio.h>
#include <string>
#include <iostream>
#include <signal.h>

// local includes =============================================================

#include "proto.h"
#include "settings.h"
#include "system.h"
#include "tools.h"

// globals ====================================================================

/// exit flag 
int canexit=0;
/// debug flag
bool debug=false;

// helper functions -----------------------------------------------------------

// cleanup on interrupt signal
void cleanup(int sig)
{
  std::cout << "rclockd shutting down." << std::endl;
  canexit=1;
  signal(SIGINT,SIG_DFL);
}

// show program help
void ShowHelp(std::string program_name)
{
  std::cout << "Usage: " << program_name << " [OPTION]... [FILE]..." 
            << std::endl << std::endl
            << "-h\t\tShows program help." << std::endl
            << "-c <file>\tSpecifies configuration file location."
            << std::endl
            << "-d\t\tRun as daemon." << std::endl
            << "-v\t\tVerbose. This will output the date command to be called."
            << std::endl;
}

/// Opens a serial connect, returns file descriptor
/// \param settings : ptr to settings object
/// \param blocking : set to false to use O_NONBLOCK 
/// \return Returns file descriptor on success (negative on failure)
int OpenSerial(c_settings* settings, bool blocking=false)
{

  // declarations
  int fd=0;
  struct termios tios;
  
  // open fd
  if (blocking == false)
  {
    fd = open(settings->GetDevice().c_str(), O_RDWR|O_NONBLOCK);
  }
  else
  {
    fd = open(settings->GetDevice().c_str(), 0);
  }

  // error check
  if(fd < 0) {
    perror("open failed: ");
    return -1;
  }

  // tcgetaddr
  if(tcgetattr(fd, &tios) < 0) 
  {
    perror("tcgetattr failed: ");
    return -1;
  }

  // ---- port setup ----
  // set 9600 baud
  cfsetspeed(&tios, B9600);  
  // set 8 bits per character
  tios.c_cflag &= ~CSIZE;
  tios.c_cflag |= CS8;
  // set no parity
  tios.c_cflag &= ~(PARENB | PARODD);
  // set one stop bit
  tios.c_cflag &= ~CSTOPB;
  // set no flow control
  tios.c_cflag &= ~CRTSCTS;
  // apply port settings
  int result = tcsetattr(fd, TCSANOW, &tios);
  if (result) { perror("setattr failed.");return 1; }

  // done, return fd
  return fd;
}

// threads --------------------------------------------------------------------

// thread to query clock
void* QueryClock(void* ptr)
{ 
  DebugOut("Write thread started.");
  // get settings object
  c_settings* settings = NULL;
  settings = static_cast<c_settings*>(ptr);
  if (!settings) return NULL;

  // open a non-blocking serial connection; get file descriptor.
  int fd = OpenSerial(settings);
  if (fd < 0) return NULL;

  // the string to send to the clock
  char str[] = "DTM?\r";

  // main loop for this thread.
  while (canexit==0)
  {
    write(fd,str,strlen(str));
    tcdrain(fd);
    DebugOut("Sent query to clock."); 
    usleep( 1000000 * settings->GetDelay() );
  }
  // close file descriptor before exit.
  close(fd);
  DebugOut("Write thread stopped.");
  return NULL;
}

// thread to read from the clock
void* ReadClock(void* ptr)
{
  DebugOut("Read thread started.");
  // get settings object
  c_settings* settings = NULL;
  settings = static_cast<c_settings*>(ptr);
  if (!settings) return NULL;

  // open a blocking serial connection; get file descriptor.
  int fd = OpenSerial(settings,true);
  if (fd < 0) return NULL;

  // declarations
  char byte = 0;
  string buffer = "";

  // main loop for this thread
  while(canexit==0)
  { 

    // init byte to 0
    byte = 0;
    
    // read one byte
    read(fd,&byte,1);
 
    // got anything?
    if(byte != 0)
    {    
      // got CR ?
      if (byte == 10)
      {
        // process buffer.
        string::size_type loc= buffer.find("DTM",0);
        if (loc != string::npos)
        {
          DebugOut("Received: " + buffer);
          settings->GetProto()->ProcessLine(buffer);
        }
        // flush buffer
        //tcflush(fd,TCIOFLUSH);
        buffer = "";
      }
      else
      {
        // feed the buffer.
        buffer += byte;      
      }
    }
  }

  // close fd on exit
  close(fd);
  DebugOut("Read thread stopped.");
  return NULL;
}

// main program ===============================================================

int main(int argc, char *argv[])
{
 
  //
  // declarations
  //
  
  using namespace std;
  // serial IO object
  struct termios tios;

  //
  // Initialisations
  //

  // who am i?
  string program_name = argv[0];
  // default config file
  string config_file = "/etc/rclockd.conf";
  // this will hold the settings object.
  c_settings* settings = NULL;

  //
  // parse program arguments
  //

  int argn = 1;
  bool valid = false;
  bool daemon = false;
  bool verbose = false;
  while (argn < argc)
  {
    string arg = argv[argn];
    
    // -c option, selects configuration file.
    if (arg == "-c")
    {
      argn++;
      if (argn >= argc)
      {
        std::cerr << "Error: missing configuration filename." << std::endl;
        ShowHelp(program_name);
        return 4;
      }
      config_file = argv[argn];
      valid = true;
    }
    // -h option, shows program help
    else if (arg == "-h")
    {
      ShowHelp(program_name);
      valid = true;
      return 0;
    }
    // -d option, daemonize
    else if (arg == "-d")
    {
      daemon = true;
      valid = true;
    }
    else if (arg == "-v")
    {
      verbose = true;
      debug = true;
      valid = true;      
    }
    else if (arg == "-vv")
    {
      verbose = true;
      debug = true;
      valid = true;
    }
    // if option is invalid, complain.
    if (!valid)
    {
      std::cerr << "Invalid option: " << argv[argn] << std::endl;
      ShowHelp(program_name);     
    }
    argn++;
  }
  // no args is valid too.  
  if (argc == 1) 
    valid = true;                                   
  // exit if there was some other problem with arguments.
  if (!valid)
    return 1;

  //
  // Load configuration file
  //

  // all ok, load config.
  settings = new c_settings(config_file);
  int load_result = settings->Load();
  if (load_result < 0)
  {      
    // settings file could not be opened
    std::cerr << "Could not open the configuration file: "
              << config_file << std::endl
              << "Please check if it exists. You can specify a" 
              << "location for the configuration file using the"
              << " -c option." << std::endl;
    return 2;
  }
  else if (load_result > 0)
  {
    std::cerr << "Error in the configuration file " << config_file
              << " on line " << (load_result+1) 
              << "." << std::endl << "Aborted." << std::endl;
    // config file parse problem.
    return 3;
  }
  
  //
  // daemonize if needed
  //
  
  c_system system(settings);
  if (daemon == true)
    system.Daemonize();
  
  //
  // apply settings to protocol handler
  //

  settings->GetProto()->SetDelay(settings->GetDelay());
  settings->GetProto()->SetDateProg(settings->GetDateProg());  
  settings->GetProto()->SetVerbose(settings->GetLogVerbose());
  if (debug == true)
    settings->GetProto()->SetVerbose(true);
 
  //
  // signal handler setup
  //

  // catch interrupts to clean up 
  signal (SIGINT,cleanup);

  // 
  // setup threads
  //

  pthread_t readThread,writeThread;
  pthread_create( &writeThread, NULL, QueryClock, static_cast<void*>(settings) );
  pthread_create( &readThread, NULL, ReadClock, static_cast<void*>(settings) );
  pthread_join( writeThread, NULL );
  pthread_join( readThread, NULL );
  return 0;
}


